package equality;

import reusable.AssociationAspect;
import reusable.IRelation;

/**
 * @author Eric Bodden
 *
 * Aspect implementing the role of a (nontransitive) equality relation.
 */
public aspect EqualityAspect extends AssociationAspect perthis(pureinstantiation()) {
    
    /* handle to left and right bit for the instance of the equality relation */
    protected Bit left, right;
    
    /* register handles (on call to associate())
     * (automatically called by the base aspect when
     *  relation is being enabled)
     */
    @Override
    protected void register(Object... keys) {
        this.left = (Bit)this.associatedObjects()[0];
        this.right = (Bit)this.associatedObjects()[1];
    }

    /* have to use the inner aspect to circumvent implicit "&& this(..)" in perthis-aspect */ 
    static aspect Inner extends AssociationAspect.SemanticPart {
        
        /* propagate set() event on left Bit to the right Bit */ 
        after(Bit b): execution(* Bit.set()) && this(b) &&!cflow(adviceexecution()){
            /* over all instances, where b is the left bit */
            for(AssociationAspect aa: aspectsOf(b,null)) {
                /* have to cast down here :-(
                 * could maybe be solved by a type paramater; but since this aspect is static,
                 * it can have no type parameter :-( */
                EqualityAspect eqa = (EqualityAspect) aa;
                /* actual propagation */
                eqa.right.set();
            }
        }
        
        /* same for other three cases */
        
        after(Bit b): execution(* Bit.set()) && this(b) &&!cflow(adviceexecution()) {
            for(AssociationAspect aa: aspectsOf(new Object[]{null,b}/* array necessary due to current JDT bug */)) {
                EqualityAspect eqa = (EqualityAspect) aa;
                eqa.left.set();
            }
        }   
        
        after(Bit b): execution(* Bit.clear()) && this(b) &&!cflow(adviceexecution()){
            for(AssociationAspect aa: aspectsOf(b,null)) {
                EqualityAspect eqa = (EqualityAspect) aa;
                eqa.right.clear();
            }
        }
        
        after(Bit b): execution(* Bit.clear()) && this(b) &&!cflow(adviceexecution()) {
            for(AssociationAspect aa: aspectsOf(new Object[]{null,b/* array necessary due to current JDT bug */})) {
                EqualityAspect eqa = (EqualityAspect) aa;
                eqa.left.clear();
            }
        }   

        /* any concrete aspect relation has an appropriate contained type; have to conretize this type here */        
        protected Class<? extends IRelation> containerType() {
            return EqualityRelation.class;
        }
    }
	
}
