package reusable;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * @author Eric Bodden
 *
 * AbstractRelation an abstract relational container.
 */
public abstract class AbstractRelation implements IRelation {

	protected Object[] keys;
	
	private static Set<AbstractRelation> vectorKeys;

	/**
     * Creates a new relation for those objects.
	 * @param relatedObjects
	 */
	public AbstractRelation(Object... relatedObjects) {
		super();
		this.keys = relatedObjects;
	}

	/**
	 * {@inheritDoc}
	 */
	public boolean equals(Object oth) {
		if (this == oth) {
			return true;
		}
	
		if (oth == null) {
			return false;
		}
	
		if (oth.getClass() != getClass()) {
			return false;
		}
	
		AbstractRelation other = (AbstractRelation) oth;
		if (!equalsHelper(keys, other.keys)) {
			return false;
		}
	
		return true;
	}
	
	/*
	 * Helper method to compare two arbitrary arrays.
	 */
	private boolean equalsHelper(Object a, Object b) {
		if (a == b) {
			return true;
		}
		if (a == null || b == null) {
			return false;
		}
	
		if (!a.getClass().isArray() || !b.getClass().isArray()) {
			return a.equals(b);
		}
	
		int aLength = java.lang.reflect.Array.getLength(a);
		if (aLength != java.lang.reflect.Array.getLength(b)) {
			return false;
		}
	
		for (int i = 0; i < aLength; i++) {
			if (!equalsHelper(
				java.lang.reflect.Array.get(a, i),
				java.lang.reflect.Array.get(b, i))) {
				return false;
			}
		}
	
		return true;
	}

	/**
	 * {@inheritDoc}
	 */
	public int hashCode() {
		final int PRIME = 1000003;
		int result = 0;
		result = PRIME * result + hashCodeHelper(keys);
	
		return result;
	}
	
	/*
	 * Helper method to compute the hashCode of arbitrary arrays.
	 */
	private int hashCodeHelper(Object a) {
		final int PRIME = 1000003;
		if (a == null) {
			return 0;
		}
		if (!a.getClass().isArray()) {
			return a.hashCode();
		}
	
		int result = 0;
		int aLength = java.lang.reflect.Array.getLength(a);
		for (int i = 0; i < aLength; i++) {
			result = PRIME
				* result
				+ hashCodeHelper(java.lang.reflect.Array.get(a, i));
		}
	
		return result;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void enable() {};
	
    /**
     * {@inheritDoc}
     */
    public void disable() {};

    /**
	 * {@inheritDoc}
	 */
	public Object[] content() {
		return this.keys.clone();
	}
	
    /**
     * @author Eric
     * Pooling for relations.
     */
    private static aspect RelationPoolingAspect {

        //does this make sense?
        declare precedence: RelationPoolingAspect, AssociationAspect;    
        
        private Map<IRelation,IRelation> relations = new HashMap<IRelation,IRelation>();
        
        pointcut p(): call(IRelation+.new(..));
        
        Object around(): p() {
            IRelation newRelation = (IRelation)proceed();
            
            if(relations.containsKey(newRelation)) {
                System.out.println("cache hit");
                return relations.get(newRelation);
            } else {
                System.out.println("cache miss");
                relations.put(newRelation,newRelation);
                return newRelation;
            }
        }
        
    }

	
}
