package reusable;

import java.util.ArrayList;
import java.util.Collection;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;

public abstract aspect AssociationAspect perthis(pureinstantiation()) {
	
	//pointcut used for perthis
    protected pointcut pureinstantiation(): execution(* IRelation+.enable());
    
    //pointcut used for aspect instantiation and association
    protected pointcut instantiation(IRelation key): pureinstantiation() && this(key);
	
    //pointcut used for disabling the relation
    protected pointcut disable(IRelation key): execution(* IRelation+.disable()) && this(key);

    //map mapping from a container type to a list of associated aspect instances
	private static Map<Class<? extends IRelation>,List<AssociationAspect>> aspectsPerAssociationType =
		new IdentityHashMap<Class<? extends IRelation>,List<AssociationAspect>>();
	
	/* the associated container for this instance of the relation */
	private IRelation relationContainer;
	
    /* return a copy of the associated objects */
	public Object[] associatedObjects() {
		return relationContainer.content().clone();
	}
	
	/**
     * Store the instance of this aspect into the static map for the appropriate container type.
	 * @param container The container this aspect instance is associated with
	 */
    @SuppressWarnings("Type safety")
	after(IRelation container): instantiation(container) {
		List<AssociationAspect> aspectsForThisRelationContainerType =
			aspectsPerAssociationType.get((Class<? extends IRelation>)container.getClass());
		if(aspectsForThisRelationContainerType==null) {
			aspectsForThisRelationContainerType = new ArrayList<AssociationAspect>();
            aspectsPerAssociationType.put((Class<? extends IRelation>)container.getClass(),aspectsForThisRelationContainerType);
		}
		if(!aspectsForThisRelationContainerType.contains(this)) {
            aspectsForThisRelationContainerType.add(this);
        }
		this.relationContainer = container;
		register(container.content());
	}
	
    @SuppressWarnings("Type safety")
    after(IRelation container): disable(container) {
        List<AssociationAspect> aspectsForThisRelationContainerType =
            aspectsPerAssociationType.get((Class<? extends IRelation>)container.getClass());
        if(aspectsForThisRelationContainerType!=null) {
            for(AssociationAspect aa: aspectsForThisRelationContainerType) {
                if(aa.relationContainer == container) {
                    aspectsForThisRelationContainerType.remove(aa);
                    break;
                }
            }
        }        
    }
    
    /**
     * This can be overriden in order to get a handle to the associated objects on instantiation.
	 * @param relatedObjects Objects bound by the relation
	 */
	protected void register(Object... relatedObjects) {}
	
	/**
	 * @author Eric Bodden
	 *
	 * SemanticPart: Semantic part of the relation implementing the functionality of the relation.
     * Here we provide only helper methods. Subaspects should implement the actual functionality.
     * Have to use a static aspect here in order to get rid of the implicit "&& this(..)" which exists
     * in the upper perthis-aspect. 
	 */
	public static abstract aspect SemanticPart {
		
        /**
         * Returns the relational container type that is used in conjunction with the implementation of the relation.
         */
        protected abstract Class<? extends IRelation> containerType();
        
		/**
         * Returns a list of all aspect instances associated with the objects <code>relatedObjects</code>.
         * Any of the positions can be <code>null</code>, being a wildcart for this position.
		 * @param relatedObjects
		 */
		protected Collection<AssociationAspect> aspectsOf(Object... relatedObjects) {
            Class<? extends IRelation> associationType = containerType();
			List<AssociationAspect> result = new ArrayList<AssociationAspect>();
            /* the following implementation may be speeded up by some more sophisticated
             * mapping technique (sorted TreeMaps for example?) */
			for(AssociationAspect eqa: aspectsPerAssociationType.get(associationType)) {
				int i=0;
				boolean equal=true;
				for(Object key: eqa.associatedObjects()) {
					if(relatedObjects[i]!=null && relatedObjects[i]!=key) {
						equal=false;
						break;
					}
					i++;
				}
				if(equal) {
					result.add(eqa);
				}
			}			
			return result;
		}
        
        /**
         * Convenience method joining two sets of aspect instances.
         * @param l1
         * @param l2
         */
        protected Collection<AssociationAspect> join(Collection<AssociationAspect> l1, Collection<AssociationAspect> l2) {
            Collection<AssociationAspect> join = new ArrayList<AssociationAspect>();
            join.addAll(l1);
            join.addAll(l2);
            return join;
        }
		
	}
	
}
